@@ -10,7 +10,7 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: cipe.h,v 1.12.4.2 1998/11/14 21:14:16 olaf Exp $ */
+/* $Id: cipe.h,v 1.19.2.1 1999/04/27 21:53:21 olaf Exp $ */
 
 #ifndef _CIPE_H_
 #define _CIPE_H_
@@ -75,6 +75,15 @@ struct  siocsifcipatt {
     int fd;
 };
 
+/* Allocate/deallocate a device. */
+#define SIOCSIFCIPALL   (SIOCDEVPRIVATE+4)
+#define SIOCSIFCIPUNA   (SIOCDEVPRIVATE+5)
+struct  siocsifcipall {
+    unsigned long   magic;
+    int num;
+    char name[16];
+};
+
 
 /*** Key exchange related definitions ***/
 
@@ -82,16 +91,15 @@ struct  siocsifcipatt {
 #define KEYXCHGBLKMIN    64
 /* Maximum kxc block, padded with random bytes */
 #define KEYXCHGBLKMAX    (KEYXCHGBLKMIN+256)
-
-/* Type words. Only 8 are possible. */
-#define TW_DATA         0x00
-#define TW_NEWKEY       0x02
-#define TW_CTRL         0x04
-#define TW_RSVD2        0x06
-#define TW_CTLBITS      0x06 /* Bit mask for all of above */
-
+/* Position of the timestamp */
+#define KEYXCHGTSPOS     56
+/* Type words. Only 4 are possible. */
+#define TW_DATA         0
+#define TW_NEWKEY       2
+#define TW_CTRL         4
+#define TW_RSVD2        6
 /* error indication, no valid type word */
-#define TW_ERROR        0x01 
+#define TW_ERROR        1
 
 /* NEWKEY (key exchange mode 1) subtypes. */
 #define NK_RREQ         0 /* not used in protocol */
@@ -99,6 +107,12 @@ struct  siocsifcipatt {
 #define NK_IND          2 /* this is my new key   */
 #define NK_ACK          3 /* i have your new key  */
 
+/* CTRL subtypes. By now sent in a TW_NEWKEY packet. */
+#define CT_DUMMY     0x70 /* ignore */
+#define CT_DEBUG     0x71 /* log */
+#define CT_PING      0x72 /* send PONG */
+#define CT_PONG      0x73
+#define CT_KILL      0x74 /* exit */
 
 /*** Kernel-module internal stuff ***/
 
@@ -111,13 +125,22 @@ struct  siocsifcipatt {
 #include <linux/sched.h>
 #include <linux/if_ether.h>
 #include <linux/net.h>
-#include <linux/etherdevice.h>
 #include <linux/ip.h>
 #include <linux/udp.h>
 #include <net/sock.h>
+#include <linux/version.h>
+#ifndef KERNEL_VERSION
+#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
+#endif
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
+#define LINUX_21
+#endif
 
-/* We attach an ethernet header to the start of each packet */
-#define cipe_pkthdr_len (sizeof(struct ethhdr))
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,127)
+#define timeout_t unsigned long
+#else
+#define timeout_t long
+#endif
 
 /* The header we add to each packet */
 #ifdef VER_SHORT
@@ -125,17 +148,10 @@ struct  siocsifcipatt {
 #else
 #define cipehdrlen (sizeof(struct iphdr)+sizeof(struct udphdr)+blockSize)
 #endif
-
-/* In the xmit function we copy the packet to a new buffer as we do
- * not know the required space for the device we are going to be
- * sending over, and the protocol layers tend to not allocate enough
- * tail room for us. Encrypted packets also tend to be larger than the
- * unencrypted ones 
- */
-#define cipexhdrl ((cipe_pkthdr_len+15)&~15)
-
+/* ...plus a real hardware header (common case) */
+#define cipexhdrl  (cipehdrlen+((ETH_HLEN+15)&~15))
 /* max. padding at the end */
-#if ProtocolVersion == 3 || ProtocolVersion == d
+#if ProtocolVersion == 3
 #define cipefootlen 12 /* 7 bytes pad, 1 byte type, 4 bytes CRC */
 #else
 #define cipefootlen 10 /* 8 bytes pad, 2 bytes CRC */
@@ -143,9 +159,9 @@ struct  siocsifcipatt {
 
 /* A CIPE device's parameter block */
 
-#define CIPE_MAGIC 0x5387
+#define CIPE_MAGIC  (htonl(0x43495045))
 struct cipe {
-    unsigned short  magic;
+    __u32           magic;
     struct device   *dev;
      /* Set by user process */
     __u32           peeraddr;
@@ -179,9 +195,17 @@ struct cipe {
     char            havekey;
     char            haveskey;
     char            haverkey;
-     /* Statistics */
+#ifdef LINUX_21
+    char            recursion;
+#endif
+    pid_t           owner;
+    /* Statistics */
+#ifdef LINUX_21
+    struct net_device_stats stat;
+#else
     struct enet_statistics stat;
-     /* Socket interface stuff */
+#endif
+    /* Socket interface stuff */
     struct proto    *udp_prot;
     struct proto    cipe_proto;
 };
@@ -195,7 +219,7 @@ struct cipe {
 
 /* Master control struct */
 struct cipe_ctrl {
-    char             name[8];
+    char             name[16];
     struct cipe      cipe;
     struct device    dev;
 };
@@ -219,12 +243,16 @@ extern int cipe_debug;
 #define DEB_OUT    4
 #define DEB_CRYPT  8
 #define DEB_KXC   16
+#define DEB_PKIN  32
+#define DEB_PKOU  32
+#define dprintk(l,p)            if(cipe_debug&l){printk p;}
 #define dprintk1(l,s,f,a)           if(cipe_debug&l){printk(s f,a);}
 #define dprintk2(l,s,f,a,b)         if(cipe_debug&l){printk(s f,a,b);}
 #define dprintk3(l,s,f,a,b,c)       if(cipe_debug&l){printk(s f,a,b,c);}
 #define dprintk4(l,s,f,a,b,c,d)     if(cipe_debug&l){printk(s f,a,b,c,d);}
 #define dprintk5(l,s,f,a,b,c,d,e)   if(cipe_debug&l){printk(s f,a,b,c,d,e);}
 #else
+#define dprintk(l,p)
 #define dprintk1(l,s,f,a)
 #define dprintk2(l,s,f,a,b)
 #define dprintk3(l,s,f,a,b,c)
@@ -233,6 +261,9 @@ extern int cipe_debug;
 #endif /* DEBUG */
 
 /* internal routines */
+/* module.c */
+extern void cipe_use_module(void);
+extern void cipe_unuse_module(void);
 /* device.c */
 extern void cipe_prnpad(unsigned char *buf, int len);
 extern void cipe_close(struct cipe *c);
@@ -240,6 +271,9 @@ extern void cipe_close(struct cipe *c);
 extern int cipe_attach(struct device *dev, struct siocsifcipatt *parm);
 void cipe_fakenkey(struct cipe *c, char typ);
 /* output.c */
+#ifdef DEBUG
+void cipe_dump_packet(char *title, struct sk_buff *skb);
+#endif
 extern int cipe_xmit(struct sk_buff *skb, struct device *dev);
 /* encaps.c */
 extern void cipe_encrypt(struct cipe *c, unsigned char *buf,
@@ -262,10 +296,6 @@ extern unsigned short block_crc(unsigned char *d, int len);
 
 #define MIN(a,b) (((a)<(b))?(a):(b))
 
-#ifndef CTLDIR
-#define CTLDIR "/etc/cipe/"
-#endif
-
 #ifndef DEVNAME
 #define DEVNAME "cip" VERNAME CRNAME
 #endif
