@@ -10,37 +10,36 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: ciped.c,v 1.16.6.1 1998/11/14 20:17:21 olaf Exp $ */
+/* $Id: ciped.c,v 1.22.2.3 1999/06/07 19:08:24 olaf Exp $ */
 
+#define CTLDIR "/etc/cipe"
+
+#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <netdb.h>
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
+#define __USE_GNU
 #include <string.h>
 #include <syslog.h>
 #include <unistd.h>
-#include <ctype.h>
-#include <mntent.h>
 #include <arpa/inet.h>
 #include <net/if.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
-#include <sys/signal.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/wait.h>
-#ifdef __GLIBC__      /* Ugh! the ehthernet address function needs this */
-#include <net/if_arp.h>
-#include <netinet/if_ether.h>
-#else
-#include <linux/if_ether.h>
-#include <linux/if_arp.h>
-#endif
-#include "linux/ipx.h"
+#include <values.h>
 #include "ciped.h"
 #include "ioctl.h"
 
+#ifndef LOGFAC
+#define LOGFAC LOG_DAEMON
+#endif
+
 static const char daemon_version[]=VERSION;
 
 void usage(const char *p)
@@ -61,7 +60,6 @@ struct options {
 	int ovint;
 	char *ovstr;
 	struct sockaddr_in *ovsaddr;
-	struct sockaddr *ovhwaddr;
     } ov;
 } opts[] = {
 #define Opt(x,t) { #x, T##t, {0} },
@@ -71,11 +69,10 @@ struct options {
 #define Taddr	3
 #define Tuaddr	4
 #define Ttaddr	5
-#define Thwaddr 6
-#define Tipxnet 7
 
-#define Odevice		0  /* name of the CIPE device, e.g. cip3b0 */
-Opt(device,str)
+#define Okey		0  /* The interface secret key
+			      - set via options file only */
+Opt(key,str)
 #define Odebug		1  /* Debug mode - don't detach, verbose messages */
 Opt(debug,bool)
 #define Oipaddr		2  /* This CIPE device's IP address */
@@ -91,9 +88,8 @@ Opt(me,uaddr)
 #define Opeer		7  /* The peer's UDP address,
 			      i.e. where to send packets */
 Opt(peer,uaddr)
-#define Okey		8  /* The interface secret key
-			      - set via options file only */
-Opt(key,str)
+#define Odevice		8  /* name of the CIPE device, e.g. cip3b0 */
+Opt(device,str)
 #define Onokey		9  /* Don't encrypt */
 Opt(nokey,bool)
 #define Osocks		10  /* SOCKS5 server (ip:port) */
@@ -112,18 +108,12 @@ Opt(arg,str)
 Opt(cttl,int)
 #define Omaxerr		17  /* Maximum ECONNREFUSED tolerance */
 Opt(maxerr,int)
-#define Onetmask        18  /* The device netmask address */
-Opt(netmask,addr)
-#define Ohwaddr         19  /* The device hardware address */
-Opt(hwaddr,hwaddr)
-#define Obroadcast      20  /* The device broadcast address */
-Opt(broadcast,addr)
-#define Obrdcstmode     21  /* An address for a static arp  for peer */
-Opt(brdcstmode,bool)       
-#define Oarphwaddr      22  /* the hardware address for the far end */
-Opt(arphwaddr,hwaddr)      
-#define Oipxnet         23  /* the CIPE links' IPX network address */
-Opt(ipxnet,ipxnet)
+#define Otokxts         18  /* Timeout for key exchange timestamps */
+Opt(tokxts,int)
+#define Oping           19  /* Keepalive ping interval */
+Opt(ping,int)
+#define Otoping         20  /* Keepalive ping timeout */
+Opt(toping,int)
 
 #undef Opt
 { NULL, 0, {0} }};
@@ -133,75 +123,17 @@ Opt(ipxnet,ipxnet)
 #define OA(x)      (opts[O##x].ov.ovsaddr)
 #define OAaddr(x)  (OA(x)->sin_addr)
 #define OAport(x)  (OA(x)->sin_port)
-#define OH(x)      (opts[O##x].ov.ovhwaddr)
 
 /* Statistics */
 struct kstat {
     int rreq, req, ind, indb, ack, ackb, bogus;
 } ks = { 0, 0, 0, 0, 0, 0, 0 };
 
+char device[16] = "";
+int devnum = -1;
 
 /*** Option parser ***/
 
-/* Parse a hardware address  - taken from ifconfig */
-int gethwaddr(char *c, struct sockaddr *sap)
-{
-    unsigned char *ptr;
-    char ch, *orig;
-    int i, val;
-
-    sap->sa_family = ARPHRD_ETHER;
-    ptr = sap->sa_data;
-
-    i = 0;
-    orig = c;
-    while((*c != '\0') && (i < ETH_ALEN)) {
-        val = 0;
-        ch = *c++;
-        if (isdigit(ch)) val = ch - '0';
-	else if (ch >= 'a' && ch <= 'f') val = ch - 'a' + 10;
-	else if (ch >= 'A' && ch <= 'F') val = ch - 'A' + 10;
-	else {
-	    fprintf(stderr, "gethwaddr: '%s' invalid hardware address\n",
-		    orig);
-	    return(-1);
-        }
-        val <<= 4;
-        ch = *c++;
-        if (isdigit(ch)) val |= ch - '0';
-	else if (ch >= 'a' && ch <= 'f') val |= ch - 'a' + 10;
-	else if (ch >= 'A' && ch <= 'F') val |= ch - 'A' + 10;
-	else {
-	    fprintf(stderr, "gethwaddr: '%s' invalid hardware address\n",
-		    orig);
-	    return(-1);
-        }
-        *ptr++ = (unsigned char) (val & 0377);
-        i++;
-
-        /* We might get a semicolon here - not required. */
-        if (*c == ':') {
-	    if (i == ETH_ALEN) {
-		; /* nothing */
-	    }
-	    c++;
-        }
-    }
-
-    return(0);
-}
-
-/* Convert an IPX address */
-int getipxnet(char *c, int *ipxnet) 
-{
-    if (sscanf(c, "%x", ipxnet)==0 || *ipxnet == 0) {
-	fprintf(stderr, "getipxnet: '%s' invalid IPX network address\n", c);
-	return -1;
-    }
-
-    return 0;
-}
-
 /* Resolve a host:port address */
 int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
 {
@@ -214,7 +146,7 @@ int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
     sa->sin_addr.s_addr=INADDR_ANY;
     if (q) {
 	if (!prot) {
-	    fprintf(stderr, "getaddr: raw IP address has no port number\n");
+	    fprintf(stderr, "raw IP address has no port number\n");
 	    return -1;
 	}
 	*q++='\0';
@@ -222,17 +154,17 @@ int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
 	    if ((s=getservbyname(q, prot)))
 		sa->sin_port=s->s_port;
 	    else {
-		fprintf(stderr, "getaddr: port '%s' invalid/not found\n", q);
+		fprintf(stderr, "port '%s' invalid/not found\n", q);
 		return -1;
 	    }
 	}
     }
-    if (!inet_aton(c, &sa->sin_addr)) {
+    if ((sa->sin_addr.s_addr=inet_addr(c))==-1) {
 	if ((h=gethostbyname(c)))
 	    memcpy(&sa->sin_addr.s_addr, h->h_addr_list[0],
 		   sizeof(sa->sin_addr.s_addr));
 	else {
-	    fprintf(stderr, "getaddr: host '%s' invalid/not found\n", c);
+	    fprintf(stderr, "host '%s' invalid/not found\n", c);
 	    return -1;
 	}
     }
@@ -267,23 +199,6 @@ int setopt(const char *n, char *v)
 		    fprintf(stderr, "setopt: out of memory\n"); return -1;
 		}
 		if (getaddr(v, o->ov.ovsaddr, protos[o->otyp-Taddr])<0) {
-		    free(o->ov.ovsaddr);
-		    o->ov.ovsaddr = NULL;
-		    return -1;
-		}
-		break;
-	    case Thwaddr:
-		if (!(o->ov.ovhwaddr=malloc(sizeof(*o->ov.ovhwaddr)))) {
-		    fprintf(stderr, "setopt: out of memory\n"); return -1;
-		}
-		if (gethwaddr(v, o->ov.ovhwaddr)<0) {
-		    free(o->ov.ovhwaddr);
-		    o->ov.ovhwaddr = NULL;
-		    return -1;
-		}
-		break;
-	    case Tipxnet:
-		if(getipxnet(v, &o->ov.ovint) < 0) {
 		    return -1;
 		}
 		break;
@@ -362,72 +277,68 @@ void setopt_file(const char *n, int v)
     fclose(f);
 }
 
+/* Make a "name=value" string of option */
+char *optstr(struct options *o)
+{
+    static char buf[1024];
+    switch(o->otyp) {
+    case Tbool:
+        snprintf(buf, sizeof(buf), "%s=%s", o->oname,
+                 o->ov.ovint?"yes":"no");
+        break;
+    case Tint:
+        snprintf(buf, sizeof(buf), "%s=%d", o->oname, o->ov.ovint);
+        break;
+    case Tstr:
+        snprintf(buf, sizeof(buf), "%s=%s", o->oname,
+                 o->ov.ovstr?o->ov.ovstr:"(none)");
+        break;
+    case Taddr:
+        if (o->ov.ovsaddr)
+            snprintf(buf, sizeof(buf), "%s=%s", o->oname,
+                     inet_ntoa(o->ov.ovsaddr->sin_addr));
+        else
+            snprintf(buf, sizeof(buf), "%s=", o->oname);
+        break;
+    case Tuaddr:
+    case Ttaddr:
+        if (o->ov.ovsaddr)
+            snprintf(buf, sizeof(buf), "%s=%s:%d", o->oname,
+                     inet_ntoa(o->ov.ovsaddr->sin_addr),
+                     ntohs(o->ov.ovsaddr->sin_port));
+        else
+            snprintf(buf, sizeof(buf), "%s=", o->oname);
+        break;
+    default: fprintf(stderr, "internal error: %d\n", __LINE__);
+    }
+    return buf;
+}
+
 /* Print out all options */
 void dumpopt(void)
 {
-    char *hwaddr;
     struct options *o;
-    for (o=opts; o->oname; ++o) {
-	printf("%s ", o->oname);
-	switch(o->otyp) {
-	case Tbool: printf("%s\n", o->ov.ovint?"yes":"no"); break;
-	case Tint:  printf("%d\n", o->ov.ovint); break;
-	case Tstr:  printf("%s\n", o->ov.ovstr?o->ov.ovstr:"(none)"); break;
-	case Taddr:
-	    if (o->ov.ovsaddr)
-		printf("%s\n", inet_ntoa(o->ov.ovsaddr->sin_addr));
-	    else
-		printf("(none)\n");
-	    break;
-	case Tuaddr:
-	case Ttaddr:
-	    if (o->ov.ovsaddr)
-		printf("%s:%d\n", inet_ntoa(o->ov.ovsaddr->sin_addr),
-		       ntohs(o->ov.ovsaddr->sin_port));
-	    else
-		printf("(none)\n");
-	    break;
-	case Thwaddr:
-	    if (o->ov.ovhwaddr) {
-		hwaddr = o->ov.ovhwaddr->sa_data;
-		printf("%02X:%02X:%02X:%02X:%02X:%02X\n",
-		       hwaddr[0],
-		       hwaddr[1],
-		       hwaddr[2],
-		       hwaddr[3],
-		       hwaddr[4],
-		       hwaddr[5]);
-	    }
-	    else
-		printf("(none)\n");
-	    break;
-	case Tipxnet:
-	    if (o->ov.ovint) {
-		printf("%08X\n", o->ov.ovint);
-	    }
-	    else
-		printf("(none)\n");
-	    break;
-	default: printf("internal error: %d\n", __LINE__); break;
-	}
-    }
+    for (o=opts; o->oname; ++o)
+        printf("%s\n", optstr(o));
 }
 
 /*** Logging ***/
 
-void log(int lev, const char *s)
+void logd(int lev, const char *f, int d)
 {
-    if (OI(debug))
-	fprintf(stderr, s);
-    else
-	syslog(lev, s);
+    if (OI(debug)) {
+	fprintf(stderr, f, d);
+        fputc('\n', stderr);
+    } else
+	syslog(lev, f, d);
 }
 
-void log2(int lev, const char *f, const char *s)
+void logs(int lev, const char *f, const char *s)
 {
-    if (OI(debug))
+    if (OI(debug)) {
 	fprintf(stderr, f, s);
-    else
+        fputc('\n', stderr);
+    } else
 	syslog(lev, f, s);
 }
 
@@ -443,224 +354,42 @@ void logerr(int lev, const char *s)
 /* missing option */
 void argmiss(const char *s)
 {
-    if (OI(debug))
-	fprintf(stderr, "missing argument: %s\n", s);
-    else
-	syslog(LOG_ERR, "missing argument: %s", s);
+    logs(LOG_ERR, "missing argument: %s", s);
 }
 
 /* print if debugging */
-#define dprintf0(f)    if(OI(debug)){printf(f);}
-#define dprintf1(f,a)  if(OI(debug)){printf(f,a);}
+#define dprintf(f)    if(OI(debug)){printf f;}
 
-
-/*** ARP code ***/
-
-/* set an ARP for the far end */
-int setarp (void) 
+void dumppacket(const char *hdr, int f, const unsigned char *buf, int len)
 {
-    struct arpreq arpreq;
-/*
- * Sometime in the 1.3 series kernels, the arp request added a device name.
- */
-#include <linux/version.h>
-#if LINUX_VERSION_CODE < 66381
-    char arpreq_arp_dev[32];
-#else
-#define  arpreq_arp_dev arpreq.arp_dev
-#endif
-    int sockfd = 0;
-
-#define SET_SA_FAMILY(addr, family)                     \
-    memset ((char *) &(addr), '\0', sizeof(addr));      \
-    addr.sa_family = (family);
-
-    memset (&arpreq, '\0', sizeof(arpreq));
-    
-    if (!OA(ptpaddr)) {
-	return 0;
+    int i=999;
+    printf("%s %d", hdr, f);
+    while (len-->0) {
+        if (++i>15) {
+            printf("\n    ");
+            i=0;
+        }
+        printf("%02x ", *buf++);
     }
-
-    if (!OH(arphwaddr)) {
-	return 0;
-    }
-    
-    if ((sockfd = socket(AF_INET,SOCK_DGRAM,0)) <0) {
-	perror("setarp: socket");
-	return -1;
-    }
-
-    memcpy(&arpreq.arp_ha, OH(arphwaddr), sizeof(*OH(arphwaddr)));
-    strncpy(arpreq_arp_dev, OS(device), sizeof(arpreq_arp_dev)-1);
-    SET_SA_FAMILY(arpreq.arp_pa, AF_INET);
-    ((struct sockaddr_in *) &arpreq.arp_pa)->sin_addr = OAaddr(ptpaddr);
-    arpreq.arp_flags = ATF_PERM;
-
-    if (ioctl(sockfd, SIOCSARP, (caddr_t)&arpreq) < 0) {
-	logerr(LOG_ERR, "setarp: SIOCSARP");
-	close(sockfd);
-	return -1;
-    }
-
-    close(sockfd);
-    return 0;
+    printf("\n");
 }
 
-/*** IPX stuff ***/
-
-/* path_to_procfs - find the path to the proc file system mount point */
-
-char procpath[256];
-int ipxconfigured = 0;
-
-int path_to_procfs (void) {
-    struct mntent *mntent;
-    FILE *fp;
-  
-    fp = fopen (MOUNTED, "r");
-    if (fp != 0) {
-	mntent = getmntent (fp);
-	while (mntent != (struct mntent *) 0) {
-	    if (strcmp (mntent->mnt_type, MNTTYPE_IGNORE) != 0) {
-		if (strcmp (mntent->mnt_type, "proc") == 0) {
-		    strncpy (procpath, mntent->mnt_dir,
-			     sizeof (procpath)-10);
-		    procpath [sizeof (procpath)-10] = '\0';
-		    fclose (fp);
-		    return 1;
-		}
-	    }
-	    mntent = getmntent (fp);
-	}
-	fclose (fp);
-    }
-    return 0;
-}
-
-/* Set the IPX network number on the interface */
-int setipxnet(void) {
-    int    result = 1;
-
-    struct stat stat_buf;
-    int    skfd; 
-    struct ifreq         ifr;
-    struct sockaddr_ipx *sipx = (struct sockaddr_ipx *) &ifr.ifr_addr;
-
-    if (!OI(ipxnet)) {
-	/* Do nothing if no net number given */
-	return 0;
-    }
-
-    if (path_to_procfs()) {
-	strcat (procpath, "/net/ipx_interface");
-	if (lstat (procpath, &stat_buf) < 0) {
-	    logerr (LOG_ERR, "setipxnet: IPX support is not present in the kernel\n");
-	    return -1;
-	}
-    }
-
-    skfd = socket (AF_IPX, SOCK_DGRAM, 0);
-    if (skfd < 0) { 
-	if (errno != EIO) {
-	    logerr(LOG_DEBUG, "setipxnet: socket(AF_IPX)");
-	}
-	result = -1;
-    }
-    else {
-	memset (&ifr, '\0', sizeof (ifr));
-	strcpy (ifr.ifr_name, OS(device));
-
-	memcpy (sipx->sipx_node, OH(hwaddr)->sa_data, IPX_NODE_LEN);
-	sipx->sipx_family  = AF_IPX;
-	sipx->sipx_port    = 0;
-	sipx->sipx_network = htonl (OI(ipxnet));
-	sipx->sipx_type    = IPX_FRAME_ETHERII;
-	sipx->sipx_action  = IPX_CRTITF;
-	/*
-	 *  Set the IPX device
-	 */
-	if (ioctl(skfd, SIOCSIFADDR, (caddr_t) &ifr) < 0) {
-	    result = -1;
-	    if (errno != EEXIST) {
-		if (errno != EIO) {
-		    logerr (LOG_DEBUG,
-			    "setipxnet: ioctl(SIOCAIFADDR, CRTITF)");
-		}
-	    }
-	    else {
-		logerr (LOG_WARNING,
-			"setipxnet: ioctl(SIOCAIFADDR, CRTITF): Address already exists");
-	    }
-	}
-    
-	ipxconfigured = 1;
-
-	close (skfd);
-    }
-    return result;
-}
-
-/* Clear the IPX network information */
-int clearipxnet (void) {
-    int    result = 0;
-    int    skfd; 
-    struct ifreq         ifr;
-    struct sockaddr_ipx *sipx = (struct sockaddr_ipx *) &ifr.ifr_addr;
-
-    if( !ipxconfigured ) {
-	/* Do not attmept this if IPX is not configured */
-	return 0;
-    }
-
-    skfd = socket (AF_IPX, SOCK_DGRAM, 0);
-    if (skfd < 0){ 
-	if (errno != EIO) {
-	    logerr(LOG_DEBUG, "clearipxnet: socket(AF_IPX)");
-	}
-	result = -1;
-    }
-    else {
-	memset (&ifr, '\0', sizeof (ifr));
-	strcpy (ifr.ifr_name, OS(device));
-
-	sipx->sipx_type    = IPX_FRAME_ETHERII;
-	sipx->sipx_action  = IPX_DLTITF;
-	sipx->sipx_family  = AF_IPX;
-/*
- *  Set the IPX device
- */
-	if (ioctl(skfd, SIOCSIFADDR, (caddr_t) &ifr) < 0) {
-	    if (errno != EIO) {
-		logerr (LOG_INFO, "clearipxnet: ioctl(SIOCAIFADDR, IPX_DLTITF)");
-	    }
-	    result = -1;
-	}
-	close (skfd);
-    }
-
-    return result;
-}
-
-
-
 /*** Initialize, open and attach the CIPE device. ***/
 
 int opendev(void)
 {
     int f, i;
+#ifndef NO_DYNDEV
+    struct siocsifcipall d;
+#endif
     struct siocsifcippar p;
     struct siocsifcipkey k;
     struct siocsifcipatt a;
-    struct ifreq r,s,t;
+    struct ifreq r;
 
 #define amiss(s) { argmiss(s); goto error; }
 #define err(s)   { logerr(LOG_ERR, s); goto error; }
 
-    if (!OS(device)) {
-	argmiss("device");
-	return -1;
-    }
-
     /* Initialize, bind and connect the socket */
     if ((f=socket(AF_INET, SOCK_DGRAM, 0))<0) {
 	logerr(LOG_ERR, "opendev: socket");
@@ -689,6 +418,26 @@ int opendev(void)
 	/* not fatal */
 	logerr(LOG_NOTICE, "opendev: getsockname");
 
+#ifdef NO_DYNDEV
+    if (!OS(device))
+        amiss("device");
+#else
+    /* Allocate the device name */
+    if (OS(device)) {
+        char *x=OS(device)+strlen(OS(device))-1;
+        while (x>=OS(device) && isdigit(*x))
+            --x;
+        d.num=atoi(++x);
+    } else {
+        d.num=-1;
+    }
+    if (ioctl_alloc(f, DEVNAME "0", &d)<0)
+        err("opendev: alloc");
+    strcpy(device, d.name);
+    devnum=d.num;
+    dprintf(("Using %s index %d\n", device, devnum));
+#endif
+
     /* Set the CIPE operation parameters */
     memset(&p, 0, sizeof(p));
     if (OA(socks)) {
@@ -699,7 +448,7 @@ int opendev(void)
     p.tmo_keylife=OI(tokey);
     p.mayclear=OI(nokey);
     p.cttl=OI(cttl);
-    if (ioctl_setpar(f, OS(device), &p)<0)
+    if (ioctl_setpar(f, device, &p)<0)
 	err("opendev: setpar");
 
     /* Set the key */
@@ -717,7 +466,7 @@ int opendev(void)
 	    *c++='\0';
 	    if (i>=userKeySize) break;
 	}
-	if (ioctl_setkey(f, OS(device), &k)<0)
+	if (ioctl_setkey(f, device, &k)<0)
 	    err("opendev: setkey");
 	memset(&k, 0, sizeof(k));
     } else {
@@ -727,63 +476,21 @@ int opendev(void)
 
     /* Perform attach */
     a.fd=f;
-    if (ioctl_attach(f, OS(device), &a)<0)
+    if (ioctl_attach(f, device, &a)<0)
 	err("opendev: attach");
 
     /* Perform ifconfig */
-    memset((char *)&r, 0, sizeof(r));
-    memset((char *)&s, 0, sizeof(s));
-    memset((char *)&t, 0, sizeof(t));
-    strcpy(r.ifr_name, OS(device));
-    strcpy(s.ifr_name, OS(device));
-    if (OH(hwaddr)) {
-	memcpy(&r.ifr_hwaddr, OH(hwaddr), sizeof(*OH(hwaddr)));
-	if (ioctl(f, SIOCSIFHWADDR, &r)<0)
-	    err("opendev: SIOCSIFHWADDR");
-      
-    }
+    strcpy(r.ifr_name, device);
     if (OA(ipaddr)) {
 	memcpy(&r.ifr_addr, OA(ipaddr), sizeof(*OA(ipaddr)));
 	if (ioctl(f, SIOCSIFADDR, &r)<0)
 	    err("opendev: SIOCSIFADDR");
     } else
 	amiss("ipaddr");
-    if (!OI(brdcstmode)) {
-	/* point to point mode - MUST have a point to point address */
-	if (!OA(ptpaddr)) {
-	    amiss("ptpaddr");
-	}
-	s.ifr_flags |= IFF_POINTOPOINT;
+    if (OA(ptpaddr)) {
 	memcpy(&r.ifr_dstaddr, OA(ptpaddr), sizeof(*OA(ptpaddr)));
 	if (ioctl(f, SIOCSIFDSTADDR, &r)<0)
 	    err("opendev: SIOCSIFDSTADDR");
-	t = s;
-	if (ioctl(f, SIOCSIFFLAGS, &t)<0)
-	    err("opendev: SIOCSIFFLAGS");
-    } 
-    else {
-	/* Broadcast mode - set IFF flags */ 
-	s.ifr_flags |= IFF_BROADCAST;
-#ifdef CIPE_MULTICAST      
-	s.ifr_flags |= IFF_MULTICAST;
-#endif
-	t = s;
-	if (ioctl(f, SIOCSIFFLAGS, &t)<0)
-	    err("opendev: SIOCSIFFLAGS");
-    }
-    if (OA(broadcast)) {
-	s.ifr_flags |= IFF_BROADCAST;
-	memcpy(&r.ifr_broadaddr, OA(broadcast), sizeof(*OA(broadcast)));
-	if (ioctl(f, SIOCSIFBRDADDR, &r)<0)
-	    err("opendev: SIOCSIFBRDADDR");
-	t = s;
-	if (ioctl(f, SIOCSIFFLAGS, &t)<0)
-	    err("opendev: SIOCSIFFLAGS");
-    }
-    if (OA(netmask)) {
-	memcpy(&r.ifr_netmask, OA(netmask), sizeof(*OA(netmask)));
-	if (ioctl(f, SIOCSIFNETMASK, &r)<0)
-	    err("opendev: SIOCSIFNETMASK");
     }
     if (OI(mtu)) {
 	r.ifr_mtu=OI(mtu);
@@ -795,19 +502,13 @@ int opendev(void)
 	if (ioctl(f, SIOCSIFMETRIC, &r)<0)
 	    err("opendev: SIOCSIFMETRIC");
     }
-    s.ifr_flags |= IFF_UP|IFF_RUNNING|IFF_NOTRAILERS;
-    if (ioctl(f, SIOCSIFFLAGS, &s)<0)
+    r.ifr_flags=IFF_UP|IFF_RUNNING|IFF_NOARP|IFF_NOTRAILERS|IFF_POINTOPOINT;
+    if (ioctl(f, SIOCSIFFLAGS, &r)<0)
 	err("opendev: SIOCSIFFLAGS");
 
-    /* Not setting the ARP for the far end of the tunnel is not fatal */
-    setarp();
-
-    /* Set up IPX network */
-    setipxnet();
-
     return f;
 
-error:
+ error:
     close(f);
     return -1;
 
@@ -820,15 +521,53 @@ error:
 void closedev(int f)
 {
     struct ifreq r;
-    strcpy(r.ifr_name, OS(device));
+    strcpy(r.ifr_name, device);
     if (ioctl(f, SIOCGIFFLAGS, &r)<0)
 	logerr(LOG_ERR, "closedev: SIOCGIFFLAGS");
     r.ifr_flags&=~(IFF_UP|IFF_RUNNING);
     if (ioctl(f, SIOCSIFFLAGS, &r)<0)
 	logerr(LOG_ERR, "closedev: SIOCSIFFLAGS");
     close(f);
+#ifndef NO_DYNDEV
+    if (devnum>0) {
+        struct siocsifcipall x;
+        if ((f=socket(AF_INET, SOCK_DGRAM, 0))<0) {
+            logerr(LOG_ERR, "closedev: socket");
+            return;
+        }
+        x.num=devnum;
+        if (ioctl_unalloc(f, device, &x)<0)
+            logerr(LOG_ERR, "closedev: alloc");
+        close(f);
+    }
+#endif
 }
 
+/* Fetch the device statistics. */
+
+char *getstats(void)
+{
+    static char b[256];
+    char *p, *q;
+    FILE *f=fopen("/proc/net/dev/", "r");
+    if (!f)
+        return NULL;
+    while (fgets(b, sizeof(b), f)) {
+        for (p=b; *p==' '; ++p);
+        for (q=p; *q && *q!=':'; ++q);
+        if (!*q)
+            continue;
+        *q++='\0';
+        if (strcmp(p, device))
+            continue;
+        fclose(f);
+        return(q);
+    }
+    fclose(f);
+    return NULL;
+}
+
+
 /*** Open a SOCKS5 connection ***/
 
 int opensocks(void) {
@@ -870,7 +609,7 @@ int eofsocks(int fd) {
 	    }
 	}
     }
-    return 0;    
+    return 0;
 }
 
 /*** Signal handler ***/
@@ -879,11 +618,6 @@ sig_atomic_t gotsig=0;
 
 void sighand(int s)
 {
-    if (OI(debug)) {
-	printf("Signal %d\n", s);
-    } else {
-	syslog(LOG_NOTICE, "got signal %d", s);
-    }
     gotsig=s;
 }
 
@@ -901,6 +635,64 @@ void setsig(int sig, void (*fun)(int))
   sigaction(sig, &sa, NULL);
 }
 
+time_t pingtime=MAXINT, pongtime=MAXINT;
+
+/*** Control message handling ***/
+
+void dprintct(const char *h, char c)
+{
+    switch(c) {
+    case CT_DUMMY: dprintf(("%s CT_DUMMY\n", h));  break;
+    case CT_DEBUG: dprintf(("%s CT_DEBUG\n", h));  break;
+    case CT_PING:  dprintf(("%s CT_PING\n", h));   break;
+    case CT_PONG:  dprintf(("%s CT_PONG\n", h));   break;
+    case CT_KILL:  dprintf(("%s CT_KILL\n",h));    break;
+    default:       dprintf(("%s %d\n", h, c));
+    }
+}
+
+/* send a control */
+int sendctl(int f, const char *msg, int len)
+{
+    int e;
+    dprintct("sending", msg[0]);
+    if (OI(nokey))
+        return -1; /* unencrypted cant handle controls */
+    if ((e=send(f, msg, len, 0))<0)
+        logerr(LOG_ERR, "sendctl");
+    return e;
+}
+
+/* handle control message */
+void ctrl(int f, char *msg, int len)
+{
+    if (OI(nokey))
+        return; /* unencrypted cant handle controls */
+    dprintct("received", msg[0]);
+    switch (msg[0]) {
+    case CT_DUMMY:
+        /* ignore it */
+        break;
+    case CT_DEBUG:
+        /* Log it */
+        logs(LOG_NOTICE, "peer: %s", msg+1);
+        break;
+    case CT_PING:
+        msg[0]=CT_PONG;
+        sendctl(f, msg, len);
+        break;
+    case CT_PONG:
+        pongtime=MAXINT;
+        break;
+    case CT_KILL:
+        logs(LOG_NOTICE, "Shutting down at peer request %s", msg+1);
+        gotsig=SIGHUP;
+        break;
+    default:
+        logd(LOG_NOTICE, "unknown control %d", msg[0]);
+    }
+}
+
 /*** The key exchange procedure ***/
 
 #ifdef VER_CRC32
@@ -939,9 +731,10 @@ int errcnt;
 int kxchg(int f, int r)
 {
     int x=0, e=0;
-    unsigned char kx[KEYXCHGBLKMIN];
+    unsigned char kx[KEYXCHGBLKMIN+2];
     #define kx_typ       (kx[0])
     #define kx_nkind_key (kx+1)
+    #define kx_ts        (*(time_t *)(kx+KEYXCHGTSPOS))
 #ifdef VER_CRC32
     #define kx_nkind_crc (*(unsigned long *)(kx+1+userKeySize))
     #define kx_nkack_crc (*(unsigned long *)(kx+1))
@@ -952,7 +745,18 @@ int kxchg(int f, int r)
     static crc scrc=0;
     static unsigned char skey[userKeySize];
 
-    if (recv(f, &kx, sizeof(kx), 0)<=0) {
+/*
+    struct msghdr hdr;
+    struct iovec iov;
+
+    iov.iov_base=&kx;
+    iov.iov_len=sizeof(kx)-1;
+    hdr.msg_name=hdr.msg_control=NULL;
+    hdr.msg_iov=&iov;
+    hdr.msg_iovlen=1;
+    hdr.msg_flags=0;
+*/
+    if ((x=recv(f, kx, KEYXCHGBLKMIN, 0))<=0) {
 	if (errno==EAGAIN || errno==EINTR)
 	     /* This is normal, would be even after select */
 	    return 0;
@@ -964,10 +768,20 @@ int kxchg(int f, int r)
 	return (OI(maxerr)<0 ? 0 : --errcnt);
     }
     errcnt=OI(maxerr);
+    /* dumppacket("recvmsg", hdr.msg_flags, kx, x); */
+    if (kx[0]>=CT_DUMMY) {
+        kx[55]='\0'; /* max length of message */
+        ctrl(f, kx, x);
+        return 0;
+    }
+    if (OI(tokxts) && (abs(time(0)-kx_ts)>OI(tokxts))) {
+        dprintf(("KX: timestamp error\n"));
+        return 0;
+    }
     switch(kx_typ) {
     case NK_RREQ:
 	kx_typ=NK_REQ;
-	dprintf0("KX: [NK_RREQ] sending NK_REQ\n");
+	dprintf(("KX: [NK_RREQ] sending NK_REQ\n"));
 	x=send(f, &kx, sizeof(kx), 0); e=errno;
 	++ks.rreq;
 	break;
@@ -983,12 +797,12 @@ int kxchg(int f, int r)
 #else
 	kx_nkind_crc=(scrc=block_crc(skey, userKeySize)); /*#*/
 #endif
-	dprintf1("KX: [NK_REQ] sending NK_IND " CRCFORM "\n", scrc);
+	dprintf(("KX: [NK_REQ] sending NK_IND " CRCFORM "\n", scrc));
 	x=send(f, &kx, sizeof(kx), 0); e=errno;
 	{
 	    struct siocsifcipkey sk;
 	    sk.which=KEY_INVAL+KEY_SEND;
-	    if (ioctl_setkey(f, OS(device), &sk)<0)
+	    if (ioctl_setkey(f, device, &sk)<0)
 		logerr(LOG_ERR, "setkey");
 	    memset(&sk, 0, sizeof(sk));
 	}
@@ -1004,25 +818,25 @@ int kxchg(int f, int r)
 	    ) {
 	    struct siocsifcipkey sk;
 	    if (dupcheck(kx_nkind_crc)) {
-		dprintf0("KX: [NK_IND] duplicate\n");
+		dprintf(("KX: [NK_IND] duplicate\n"));
 		++ks.indb;
 	    } else {
 		sk.which=KEY_RECV;
 		memcpy(&sk.thekey, kx_nkind_key, userKeySize);
-		if (ioctl_setkey(f, OS(device), &sk)<0)
+		if (ioctl_setkey(f, device, &sk)<0)
 		    perror("setkey");
 		else {
 		    kx_typ=NK_ACK;
 		    kx_nkack_crc=kx_nkind_crc; /* XX */
-		    dprintf1("KX: [NK_IND] sending NK_ACK " CRCFORM "\n",
-			     (kx_nkack_crc)); /*#*/
+		    dprintf(("KX: [NK_IND] sending NK_ACK " CRCFORM "\n",
+			     (kx_nkack_crc))); /*#*/
 		    x=send(f, &kx, sizeof(kx), 0); e=errno;
 		}
 		++ks.ind;
 	    }
 	    memset(&sk, 0, sizeof(sk));
 	} else {
-	    dprintf0("KX: [NK_IND] invalid\n");
+	    dprintf(("KX: [NK_IND] invalid\n"));
 	    ++ks.indb;
 	}
 	break;
@@ -1031,25 +845,28 @@ int kxchg(int f, int r)
 	    struct siocsifcipkey sk;
 	    sk.which=KEY_SEND;
 	    memcpy(&sk.thekey, &skey, userKeySize);
-	    dprintf1("KX: [NK_ACK] got " CRCFORM "\n", scrc);
-	    if (ioctl_setkey(f, OS(device), &sk)<0)
+	    dprintf(("KX: [NK_ACK] got " CRCFORM "\n", scrc));
+	    if (ioctl_setkey(f, device, &sk)<0)
 		perror("setkey");
 	    ++ks.ack;
 	    memset(&sk, 0, sizeof(sk));
 	    memset(&skey, 0, sizeof(skey));
 	} else {
-	    dprintf1("KX: [NK_ACK] unknown " CRCFORM "\n", scrc);
+	    dprintf(("KX: [NK_ACK] unknown " CRCFORM "\n", scrc));
 	    ++ks.ackb;
 	}
 	break;
     default:
-	dprintf1("KX: invalid type %02X\n", kx_typ);
+	dprintf(("KX: invalid type %02X\n", kx_typ));
 	++ks.bogus;
     }
     memset(kx, 0, sizeof(kx));
     if (x<0) {
 	errno=e;
-	logerr(LOG_WARNING, "kxchg: send");
+	logerr(LOG_NOTICE, "kxchg: send");
+	if (e==ENXIO)
+	    return -1;
+	x=(OI(maxerr)<0 ? 0 : --errcnt);
     }
     return x;
 }
@@ -1067,22 +884,17 @@ void pspawn(char *s)
     if (c) {
 	(void) wait(NULL); /* exactly one child is active */
     } else {
-      	/* call <interface> <my-addr> <daemon-pid> <local> <remote> <network-address> <netmask> <arg>*/
-        struct in_addr networkaddr;
-	struct in_addr netmask;
-	char bufa[32], bufb[32], bufc[32], bufd[32], bufe[32], buff[32];
-	char *na[]={s, OS(device), bufa, bufb, bufc, bufd, 
-		    bufe, buff, OS(arg), NULL};
-
-	netmask.s_addr = (OA(netmask) ?  OAaddr(netmask).s_addr : 0xffffffff);
-	networkaddr.s_addr = htonl(ntohl(OAaddr(ipaddr).s_addr) 
-				   & ntohl(OAaddr(netmask).s_addr));
+	/* call <interface> <my-addr> <daemon-pid> <local> <remote> <arg> */
+	char bufa[32], bufb[32], bufc[32], bufd[32];
+	char *na[]={s, device, bufa, bufb, bufc, bufd, OS(arg), NULL};
+        struct options *o = opts;
 	sprintf(bufa, "%s:%d", inet_ntoa(OAaddr(me)), ntohs(OAport(me)));
 	sprintf(bufb, "%d", p);
 	sprintf(bufc, "%s", inet_ntoa(OAaddr(ipaddr)));
 	sprintf(bufd, "%s", inet_ntoa(OAaddr(ptpaddr)));
-	sprintf(bufe, "%s", inet_ntoa(networkaddr));
-	sprintf(buff, "%s", inet_ntoa(netmask));
+        /* put all options but the first (key) into environment */
+        for (o=opts+1; o->oname; ++o)
+            putenv(strdup(optstr(o)));
 	close(0); close(1); close(2);
 	(void) open("/dev/null", O_RDWR);
 	(void) dup(0); (void) dup(0);
@@ -1096,7 +908,9 @@ void pspawn(char *s)
 
 int main(int argc, char *argv[])
 {
-    int i, ur, fd, sd=-1;
+    int i, ur, fd, sd=-1, pi[2];
+    time_t t;
+    char *st;
     char *pn=strrchr(argv[0], '/');
     if (pn)
 	++pn;
@@ -1120,67 +934,105 @@ int main(int argc, char *argv[])
 	}
     }
     setopt_cmdline(--argc, ++argv);
+    if (!OI(maxerr))
+	OI(maxerr)=8; /* compatibility */
 
     for (i=getdtablesize()-1; i>2; --i)
 	(void) close(i);
-
     if (!OI(debug)) {
 	/* daemonize */
+        if (pipe(pi)<0) {
+            perror("pipe"); exit(1);
+        }
 	if ((i=fork())<0) {
 	    perror("fork"); exit(1);
 	}
-	if (i)
+	if (i) {
+            /* wait until the child is ready */
+            close(pi[1]);
+            (void) read(pi[0], &i, 1);
 	    exit(0);
-	openlog(pn, LOG_PID, LOG_DAEMON);
+        }
+        close(pi[0]);
+	openlog(pn, LOG_PID, LOGFAC);
 	/* don't let the syslog socket get in the way of the standard fds */
 	close(0); close(1); close(2);
 	setsid();
-	syslog(LOG_INFO, "%s: CIPE daemon vers %s (c) Olaf Titz 1996-1998",
-	       OS(device), daemon_version);
+	syslog(LOG_INFO, "CIPE daemon vers %s (c) Olaf Titz 1996-1998",
+	       daemon_version);
     } else {
-	printf("%s: CIPE daemon vers %s (c) Olaf Titz 1996-1998\n",
-	       OS(device), daemon_version);
+	printf("CIPE daemon vers %s (c) Olaf Titz 1996-1998\n",
+	       daemon_version);
 	dumpopt();
     }
 
+    setsig(SIGHUP, sighand);
+    setsig(SIGINT, sighand);
+    setsig(SIGTERM, sighand);
+    errcnt=OI(maxerr);
     if ((ur=open("/dev/urandom", O_RDONLY))<0) {
 	logerr(LOG_ERR, "open(/dev/urandom)"); exit(1);
     }
+    i=MAXINT;
     if (OA(socks)) {
-	struct itimerval t;
-	if ((sd=opensocks())<0)
+        i=30; /* set timeout for control socket check */
+        if ((sd=opensocks())<0)
 	    exit(1);
-	/* Since the key exchange socket can't do select(), we make
-	   the recv() return EINTR every now and then to get a chance
-	   to check the socks5 control socket. */
-	setsig(SIGALRM, sigignore);
-	t.it_interval.tv_sec = 30;
-	t.it_interval.tv_usec = 0;
-	t.it_value.tv_sec = 60;
-	t.it_value.tv_usec = 0;
-	setitimer(ITIMER_REAL, &t, 0);
     }
     if ((fd=opendev())<0)
 	exit(1);
     pspawn(OS(ipup) ? OS(ipup) : CTLDIR "/ip-up");
-    setsig(SIGHUP, sighand);
-    setsig(SIGINT, sighand);
-    setsig(SIGTERM, sighand);
-    if (!OI(maxerr))
-	OI(maxerr)=8; /* compatibility */
-    errcnt=OI(maxerr);
 
-    /* Main loop */
-    while ((!gotsig) && (kxchg(fd, ur)>=0)) {
-	if ((sd>=0) && (i=eofsocks(sd))) {
-	    log2(LOG_INFO, "%s on socks connection", eofsocks_str[i]);
-	    break;
-	}
+    /* Set timer for periodic checks: ping, ping-timeout */
+    if (OI(ping)) {
+        pingtime=time(0);
+        if (OI(ping)<i)
+            i=OI(ping);
+    }
+    if (OI(toping)) {
+        if (OI(toping)<i)
+            i=OI(toping);
+    }
+    if (i<MAXINT) {
+	struct itimerval t;
+	setsig(SIGALRM, sigignore);
+	t.it_interval.tv_sec = i;
+	t.it_interval.tv_usec = 0;
+	t.it_value.tv_sec = i+10;
+	t.it_value.tv_usec = 0;
+	setitimer(ITIMER_REAL, &t, 0);
     }
 
-    /* Clear ipx configuration */
-    clearipxnet();
+    if (!OI(debug))
+        close(pi[1]);
+    /* sendctl(fd, "\x71starting up", 13); */
 
+    /* Main loop */
+    while (kxchg(fd, ur)>=0) {
+        if (gotsig) {
+            logs(LOG_NOTICE, "got %s", strsignal(gotsig));
+            break;
+        }
+	if ((sd>=0) && ((i=eofsocks(sd)))) {
+	    logs(LOG_NOTICE, "%s on socks connection", eofsocks_str[i]);
+	    break;
+	}
+        t=time(0);
+        if (t>=pongtime) {
+            logs(LOG_NOTICE, "keepalive timeout", NULL);
+            if (OI(maxerr)>=0 && (--errcnt<0))
+                break;
+        }
+        if (t>=pingtime) {
+            char buf=CT_PING;
+            sendctl(fd, &buf, 1);
+            pingtime=t+OI(ping);
+            pongtime=t+OI(toping);
+        }
+    }
+    sendctl(fd, "\x71going down", 12);
+    if ((st=getstats()))
+        logs(LOG_INFO, "Interface stats %s\n", st);
     closedev(fd);
     if (OI(debug))
 	printf("KX stats: rreq=%d, req=%d, ind=%d, indb=%d, ack=%d,"
@@ -1191,7 +1043,6 @@ int main(int argc, char *argv[])
 	       " ackb=%d, unknown=%d\n", ks.rreq, ks.req, ks.ind, ks.indb,
 	       ks.ack, ks.ackb, ks.bogus);
     pspawn(OS(ipdown) ? OS(ipdown) : CTLDIR "/ip-down");
-    if (!OI(debug))
-	syslog(LOG_INFO, "%s: daemon exiting", OS(device));
+    logs(LOG_INFO, "%s: daemon exiting", device);
     return 0;
 }
