@@ -10,13 +10,13 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: ciped.h,v 1.1 1997/09/18 23:00:46 olaf Exp $ */
+/* $Id: ciped.h,v 1.2.2.1 1999/04/27 21:53:22 olaf Exp $ */
 
 #include "cipe.h"
 
 /* ciped.c */
-extern void log(int lev, const char *s);
-extern void log2(int lev, const char *f, const char *s);
+extern void logd(int lev, const char *f, int d);
+extern void logs(int lev, const char *f, const char *s);
 extern void logerr(int lev, const char *s);
 
 /* socks5.c */
