@@ -10,7 +10,7 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: crypto.h,v 1.3.4.1 1998/11/14 20:17:21 olaf Exp $ */
+/* $Id: crypto.h,v 1.5 1999/04/25 16:42:59 olaf Exp $ */
 
 #ifndef _CRYPTO_H_
 #define _CRYPTO_H_
@@ -20,17 +20,32 @@ typedef unsigned long part;
    Used only for optimizing block-copy and block-XOR operations. */
 
 #if     ProtocolVersion == 1
+
+#ifdef  OLDNAMES
 #define VERNAME "1"
+#else
+#define VERNAME "a"
+#endif
 #define VER_BACK                /* encryption progress backwards */
 #define VER_SHORT               /* no IV in packet */
+
 #elif   ProtocolVersion == 2
+
+#ifdef  OLDNAMES
 #define VERNAME "2"
+#else
+#define VERNAME "b"
+#endif
+
 #elif   ProtocolVersion == 3
+
+#ifdef  OLDNAMES
 #define VERNAME "3"
+#else
+#define VERNAME "c"
+#endif
 #define VER_CRC32               /* checksums are 32bit */
-#elif   ProtocolVersion == d
-#define VERNAME "d"
-#define VER_CRC32               /* checksums are 32bit */
+
 #else
 #error  "Must specify correct ProtocolVersion"
 #endif
@@ -62,7 +77,7 @@ typedef unsigned long part;
 #define blockSize               sizeof(Blowfish_Data)
 
 #else
-#error  "Must specify Crypto_Idea or CryptoBlowfish"
+#error  "Must specify Crypto_IDEA or CryptoBlowfish"
 #endif
 #endif
 
