@@ -10,28 +10,42 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: device.c,v 1.15.4.2 1998/11/14 21:14:16 olaf Exp $ */
+/* $Id: device.c,v 1.22 1999/04/25 16:42:59 olaf Exp $ */
 
 #include "cipe.h"
 #include "version.h"
+#include <stddef.h>
 #include <linux/if.h>
 #include <linux/if_arp.h>
-#include <linux/version.h>
+#include <linux/sched.h>
+
+#ifdef LINUX_21
+#include <asm/uaccess.h>
+#include <linux/rtnetlink.h>
+#else
+#define register_netdevice register_netdev
+#define unregister_netdevice unregister_netdev
+#endif
 
 /*** Globals ***/
 
 static const char driver_version[]=VERSION;
 
 struct cipe_ctrl **cipe_ctrls = NULL;
+#ifdef NO_DYNDEV
 int cipe_maxdev = 4;            /* changeable via insmod */
+#else
+int cipe_maxdev = 100;          /* changeable via insmod */
+#endif
 #ifdef DEBUG
 int cipe_debug = DEB_CALL;      /* changeable via insmod */
 #endif
 
-/* clear all potentially sensitive info */
+/* clear all potentially sensitive info and stats */
 static void cipe_zero_c(struct cipe *c)
 {
-    memset(&(c->peeraddr), 0, (char*)&(c->stat)-(char*)&(c->peeraddr));
+    memset(&(c->peeraddr), 0,
+           offsetof(struct cipe, udp_prot)-offsetof(struct cipe, peeraddr));
     /* reset these to sensible values */
     c->tmo_keyxchg = 10*HZ;
     c->tmo_keylife = 10*60*HZ;
@@ -86,6 +100,8 @@ static int cipe_setpar(struct device *dev, struct siocsifcippar *parm)
 	c->tmo_keylife=parm->tmo_keylife*HZ;
     c->mayclear=parm->mayclear;
     c->cttl=parm->cttl;
+    c->dev->hard_header_len=cipexhdrl+
+	((c->sockshost) ? sizeof(struct sockshdr) : 0);
     return 0;
 }
 
@@ -93,7 +109,7 @@ static int cipe_setkey(struct device *dev, struct siocsifcipkey *parm)
 {
     DEVTOCIPE(dev,c,-ENODEV);
 
-    dprintk2(DEB_KXC, KERN_INFO, "%s: setkey %d\n", dev->name, parm->which);
+    dprintk(DEB_KXC, (KERN_INFO "%s: setkey %d\n", dev->name, parm->which));
     switch (parm->which) {
     case KEY_STATIC:
 	ExpandUserKey(parm->thekey, c->key_e);
@@ -134,22 +150,142 @@ static int cipe_setkey(struct device *dev, struct siocsifcipkey *parm)
     return 0;
 }
 
+static int cipe_alloc_dev(int n);
+static void cipe_unalloc_dev(int n);
+
+#ifndef NO_DYNDEV
+static int cipe_hasowner(int n)
+{
+    struct task_struct *p=current;
+    pid_t pid=cipe_ctrls[n]->cipe.owner;
+    if (!pid)
+        return 0;
+    do {
+        if (p->pid==pid)
+            return 1;
+        p=p->next_task;
+    } while (p!=current);
+    return 0;
+}
+#endif
+
+#ifdef LINUX_21
+/* In 2.1 the ioctl operations are run under lock. Beware of deadlocks. */
+#define cipe_alloc_LOCK()       0
+#define cipe_alloc_UNLOCK()
+#else
+static struct semaphore cipe_alloc_sem=MUTEX;
+#define cipe_alloc_LOCK()       down_interruptible(&cipe_alloc_sem)
+#define cipe_alloc_UNLOCK()     up(&cipe_alloc_sem)
+#endif
+
+static int cipe_alloc(struct device *dev, struct siocsifcipall *parm)
+{
+#ifdef NO_DYNDEV
+    return -ENOSYS;
+#else
+    int n=parm->num;
+    int e;
+    if (n>=cipe_maxdev)
+        return -EINVAL;
+    if ((e=cipe_alloc_LOCK()))
+        return e;
+    if (n>=0) {
+        if (cipe_ctrls[n]) {
+            if (cipe_ctrls[n]->cipe.sock || cipe_hasowner(n))
+                e=-EBUSY;
+            else
+                cipe_ctrls[n]->cipe.owner=current->pid;
+        } else {
+            e=cipe_alloc_dev(n);
+        }
+    } else {
+        e=-EMFILE;
+        for (n=0; n<cipe_maxdev; ++n) {
+            if (!cipe_ctrls[n]) {
+                e=cipe_alloc_dev(n);
+                break;
+            }
+            if (!cipe_hasowner(n)) {
+                cipe_ctrls[n]->cipe.owner=current->pid;
+                e=0;
+                break;
+            }
+        }
+    }
+    if (!e) {
+        parm->num=n;
+        strncpy(parm->name, cipe_ctrls[n]->name, sizeof(parm->name)-1);
+        parm->name[sizeof(parm->name)-1]='\0';
+    }
+    cipe_alloc_UNLOCK();
+    return e;
+#endif
+}
+
+static int cipe_unalloc(struct device *dev, struct siocsifcipall *parm)
+{
+#ifdef NO_DYNDEV
+    return -ENOSYS;
+#else
+    int e;
+    if (parm->num<0 || parm->num>=cipe_maxdev)
+        return -EINVAL;
+    if ((e=cipe_alloc_LOCK()))
+        return e;
+    if (cipe_ctrls[parm->num]->cipe.sock) {
+        e=-EBUSY;
+    } else {
+        if (parm->num>0)
+            cipe_unalloc_dev(parm->num);
+    }
+    cipe_alloc_UNLOCK();
+    return e;
+#endif
+}
+
 
 /*** Device operation handlers ***/
 
 int cipe_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
 {
     int e;
+
+#ifdef LINUX_21
+
+    if (!capable(CAP_NET_ADMIN))
+	return -EPERM;
+
+#define doioctl(nam,fun,str) {                                          \
+    struct str parm;                                                    \
+    dprintk(DEB_CALL, (KERN_INFO "%s: " nam "\n", dev->name));          \
+    if ((e=copy_from_user((void*)&parm,(void*)ifr->ifr_data,            \
+                          sizeof(parm)))<0)                             \
+        return e;                                                       \
+    if (parm.magic!=VERSION_MAGIC) {                                    \
+        printk(KERN_WARNING "%s: ciped version mismatch", dev->name);   \
+        return -EINVAL; }                                               \
+    if ((e=fun(dev, &parm))<0)                                          \
+        return e;                                                       \
+    if ((e=copy_to_user((void*)ifr->ifr_data, (void*)&parm,             \
+                        sizeof(parm)))<0)                               \
+        return e;                                                       \
+    return 0; }
+
+#else
+
     if (!suser())
 	return -EPERM;
 
 #define doioctl(nam,fun,str) {                                              \
     struct str parm;                                                        \
-    dprintk1(DEB_CALL, KERN_INFO, "%s: " nam "\n", dev->name);              \
+    dprintk(DEB_CALL, (KERN_INFO "%s: " nam "\n", dev->name));              \
     if ((e=verify_area(VERIFY_READ, ifr->ifr_data, sizeof(parm)))<0)        \
         return e;                                                           \
     memcpy_fromfs((void*)&parm, (void*)ifr->ifr_data, sizeof(parm));        \
-    if (parm.magic!=VERSION_MAGIC) return -EINVAL;                          \
+    if (parm.magic!=VERSION_MAGIC) {                                        \
+        printk(KERN_WARNING "%s: ciped version mismatch", dev->name);       \
+        return -EINVAL; }                                                   \
     if ((e=fun(dev, &parm))<0)                                              \
         return e;                                                           \
     if ((e=verify_area(VERIFY_WRITE, ifr->ifr_data, sizeof(parm)))<0)       \
@@ -157,6 +293,8 @@ int cipe_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
     memcpy_tofs((void*)ifr->ifr_data, (void*)&parm, sizeof(parm));          \
     return 0; }
 
+#endif
+
     switch (cmd) {
     case SIOCGIFCIPPAR:
 	doioctl("getpar", cipe_getpar, siocgifcippar);
@@ -166,6 +304,10 @@ int cipe_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
 	doioctl("setkey", cipe_setkey, siocsifcipkey);
     case SIOCSIFCIPATT:
 	doioctl("attach", cipe_attach, siocsifcipatt);
+    case SIOCSIFCIPALL:
+	doioctl("alloc", cipe_alloc, siocsifcipall);
+    case SIOCSIFCIPUNA:
+	doioctl("unalloc", cipe_unalloc, siocsifcipall);
     default:
 	return -EINVAL;
     }
@@ -175,36 +317,29 @@ int cipe_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
 
 int cipe_dev_open(struct device *dev)
 {
-#if ProtocolVersion == d
-    __u8 mac0[ETH_ALEN] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
-
     DEVTOCIPE(dev,c,-ENODEV);
-    if(memcmp(mac0, dev->dev_addr, ETH_ALEN) == 0)
-	return -ENXIO;
-#else
-    DEVTOCIPE(dev,c,-ENODEV);
-#endif
     if (!c->sock)
 	return -ENXIO;
-    dprintk1(DEB_CALL, KERN_INFO, "%s: opened\n", dev->name);
+    dprintk(DEB_CALL, (KERN_INFO "%s: opened\n", dev->name));
     return 0;
 }
 
 void cipe_close(struct cipe *c)
 {
     cipe_zero_c(c);
-    dprintk1(DEB_CALL, KERN_INFO, "%s: closed\n", c->dev->name);
+    dprintk(DEB_CALL, (KERN_INFO "%s: closed\n", c->dev->name));
 }
 
 int cipe_dev_close(struct device *dev)
 {
     struct cipe *c = (struct cipe*)(dev->priv);
     if ((!c) || (c->magic!=CIPE_MAGIC)) {
-	printk(KERN_WARNING "%s: cipe_close(): no valid struct\n", dev->name);
+	printk(KERN_WARNING "%s: cipe_dev_close(): no valid struct\n",
+               dev->name);
 	return 0;
     }
     if (c->sock) {
-	dprintk1(DEB_CALL, KERN_INFO, "%s: closing\n", c->dev->name);
+	dprintk(DEB_CALL, (KERN_INFO "%s: closing\n", c->dev->name));
 	/* Tell the attached socket we're going down */
 	c->sock->shutdown=SHUTDOWN_MASK;
 	c->sock->zapped=1;
@@ -222,92 +357,21 @@ struct enet_statistics *cipe_get_stats(struct device *dev)
     return &(c->stat);
 }
 
-#if ProtocolVersion == d
-/*** Frame Header Routines - stolen from net/ethernet/eth.c ***/
-
-
-/* Multi-cast function, just a blank placeholder */
-
-static void
-cipe_set_multicast_list (struct device *dev) {
-
-    if (dev->flags&IFF_PROMISC) {
-    }
-
-    else if((dev->flags&IFF_ALLMULTI)) {
-    }
-  
-    else if(dev->mc_count) {
-    }
-  
-    else {
-    }
-}
-
 
 /*** Initialization and finalization stuff ***/
 
-static int cipe_init_dev(struct device *dev)
-{
-    __u8 bmac[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
-    __u8 smac[ETH_ALEN] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
-    struct device *eth0; 
-    struct cipe *c = (struct cipe*)(dev->priv);
-
-    if (!c)
-	return -ENODEV;
-
-    memset(c, 0, sizeof(struct cipe)); /* zero the device struct along */
-    c->magic       = CIPE_MAGIC;
-    c->dev         = dev;
-    cipe_zero_c(c);
-
-    /* Find the address of ethernet 0, if possible and fill in our
-     * device source address with it, this is a bit naughty, but 
-     * it gives us a unique MAC address
-     */
-    eth0 = dev_get("eth0");
-
-    if( eth0 && memcmp(eth0->dev_addr, smac, ETH_ALEN) != 0 ) {
-	memcpy(smac, eth0->dev_addr, ETH_ALEN);
-    }
-
-
-    /* Call ether_setup to initialise most of the stuff - we will change
-     * what we need to afterwards 
-     */
-    ether_setup(dev);
-
-    /* Device parameters. */
-    /* Procedural */
-    dev->open                   = cipe_dev_open;
-    dev->stop                   = cipe_dev_close;
-    dev->hard_start_xmit        = cipe_xmit;
-    dev->get_stats              = cipe_get_stats;
-    dev->do_ioctl               = cipe_dev_ioctl;
-    dev->set_multicast_list     = cipe_set_multicast_list;
-
-    /* "Hardware" */
-    /* dev->hard_header_len 	= cipe_pkthdr_len; */
-    dev->mtu		        = ETH_DATA_LEN
-	                          -cipehdrlen
-	                          -cipefootlen
-	                          -cipe_pkthdr_len
-	                          -sizeof(struct sockshdr);
-    memcpy(dev->broadcast, bmac, ETH_ALEN);
-    memcpy(dev->dev_addr, smac, ETH_ALEN);
-
-    /* New-style flags */
-    dev->flags		|= IFF_NOTRAILERS;
-    dev->metric         = 1;
-    
-    return 0;
-}
-
-#else /* if Protocol Version == d */
-static int cipe_init_dev(struct device *dev)
+#ifndef LINUX_21
+static inline void dev_init_buffers(struct device *dev)
 {
     int i;
+    for (i = 0; i < DEV_NUMBUFFS; i++)  {
+        skb_queue_head_init(&dev->buffs[i]);
+    }
+}
+#endif
+
+static int cipe_init_dev(struct device *dev)
+{
     struct cipe *c = (struct cipe*)(dev->priv);
     if (!c)
 	return -ENODEV;
@@ -330,33 +394,39 @@ static int cipe_init_dev(struct device *dev)
     /* "Hardware" */
     dev->type		        = ARPHRD_TUNNEL;
     dev->hard_header_len 	= cipexhdrl;
-    dev->mtu		        = ETH_DATA_LEN-cipehdrlen-
-	                             cipefootlen-sizeof(struct sockshdr);
+    dev->mtu		        = ETH_DATA_LEN
+                                     -sizeof(struct sockshdr)
+                                     -cipehdrlen
+                                     -cipefootlen;
+
     dev->addr_len		= 4;            /* Dummy? */
     dev->tx_queue_len	        = 100; /* matches ethernet */
 
-    for (i = 0; i < DEV_NUMBUFFS; i++)  {
-	skb_queue_head_init(&dev->buffs[i]);
-    }
-
-    /* New-style flags */
-    dev->flags		= IFF_POINTOPOINT|IFF_NOTRAILERS|IFF_NOARP;
+#ifdef LINUX_21
+    dev->iflink         = -1;
+#else
     dev->family		= AF_INET;
     dev->pa_alen	= 4;
     dev->metric         = 1;
+#endif
+    dev_init_buffers(dev);
+
+    /* New-style flags */
+    dev->flags		= IFF_POINTOPOINT|IFF_NOTRAILERS|IFF_NOARP;
 
     return 0;
 }
 
-#endif
-
-
-static int cipe_alloc(int n)
+static int cipe_alloc_dev(int n)
 {
-    struct cipe_ctrl *cc = kmalloc(sizeof(struct cipe_ctrl), GFP_KERNEL);
-    if (!cc)
-	return -ENOMEM;
-    cipe_ctrls[n] = cc;
+    int e=0;
+    struct cipe_ctrl *cc;
+
+    dprintk(DEB_CALL, (KERN_INFO DEVNAME ": cipe_alloc_dev %d\n", n));
+    if (!(cc=kmalloc(sizeof(struct cipe_ctrl), GFP_KERNEL))) {
+        cipe_ctrls[n]=NULL;
+        return -ENOMEM;
+    }
 
     memset((void *)cc, 0, sizeof(struct cipe_ctrl));
     sprintf(cc->name, DEVNAME "%d", n);
@@ -364,36 +434,42 @@ static int cipe_alloc(int n)
     cc->dev.base_addr = n; /* dummy */
     cc->dev.priv      = (void*)&(cc->cipe);
     cc->dev.next      = NULL;
-    cc->dev.init      = cipe_init_dev; /* called by register_netdev */
+    cc->dev.init      = cipe_init_dev; /* called by register_netdevice */
 
-    if (register_netdev(&(cc->dev))<0) {
+    e=register_netdevice(&(cc->dev));
+    if (e<0) {
 	kfree(cc);
-	cipe_ctrls[n]=NULL;
 	printk(KERN_ERR
-	       "%s: register_netdev() failed\n", cc->name);
+	       "%s: register_netdevice() failed\n", cc->name);
+        cc=NULL;
+    } else {
+        cc->cipe.owner=current->pid;
     }
-    return 0;
+    cipe_ctrls[n]=cc;
+    return e;
 }
 
-static void cipe_unalloc(int n)
+static void cipe_unalloc_dev(int n)
 {
     struct cipe_ctrl *cc=cipe_ctrls[n];
     if (!cc)
 	return;
-    unregister_netdev(&(cc->dev));
-    kfree(cc);
+    dprintk(DEB_CALL, (KERN_INFO DEVNAME ": cipe_unalloc_dev %d\n", n));
+    if (cc->cipe.magic!=CIPE_MAGIC) {
+        printk(KERN_WARNING DEVNAME ": Ouch: cipe_unalloc_dev() wrong struct\n");
+        return;
+    }
+    unregister_netdevice(&(cc->dev));
     cipe_ctrls[n]=NULL;
+    kfree(cc);
 }
 
-#ifdef MODULE
 
 int init_module(void)
 {
-    int i;
-
-     /* sanity check on insmod-provided data */
+    /* sanity check on insmod-provided data */
     if (cipe_maxdev<1)  cipe_maxdev=1;
-    if (cipe_maxdev>99) cipe_maxdev=99;
+    if (cipe_maxdev>100) cipe_maxdev=100;
 
 #ifdef DEBUG
     printk(KERN_INFO
@@ -414,32 +490,24 @@ int init_module(void)
 	return -ENOMEM;
     }
     memset(cipe_ctrls, 0, sizeof(void*)*cipe_maxdev);
-    for (i=0; i<cipe_maxdev; ++i)
-	if (cipe_alloc(i)<0)
-	    return -ENOMEM;
-    return 0;
+#ifdef NO_DYNDEV
+    {
+        int i, e;
+        for (i=0; i<cipe_maxdev; ++i)
+            if ((e=cipe_alloc_dev(i)))
+                return e;
+        return 0;
+    }
+#else
+    return cipe_alloc_dev(0);
+#endif
 }
 
 void cleanup_module(void)
 {
     int i;
     for (i=0; i<cipe_maxdev; ++i)
-	cipe_unalloc(i);
+	cipe_unalloc_dev(i);
     kfree(cipe_ctrls);
 }
 
-#else
-
-#error "This does only work as a module."
-
-#endif /* MODULE */
-
-#if LINUX_VERSION_CODE < 131102 /* < 2.0.30 */
-/* older kernel not always exported this */
-
-int bad_user_access_length(void)
-{
-    panic("bad_user_access_length in" DEVNAME);
-}
-
-#endif
