@@ -10,7 +10,7 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: encaps.c,v 1.9.4.1 1998/11/14 20:17:22 olaf Exp $ */
+/* $Id: encaps.c,v 1.11.2.1 1999/04/27 21:53:22 olaf Exp $ */
 
 #include "cipe.h"
 #include <string.h>
@@ -119,15 +119,15 @@ void cipe_cryptpad(unsigned char *buf)
 
     if (++padcnt>MAXBLKS) {
 	/* make a new random key */
-	char k[userKeySize];
+	UserKey k;
 	dprintk1(DEB_CRYPT, KERN_INFO, "%s: re-keying cryptpad\n", DEVNAME);
-	cipe_prnpad(k, userKeySize);
+	cipe_prnpad((unsigned char*)k, sizeof(UserKey));
 	ExpandUserKey(k, padkey);
 	padcnt=0;
     }
     *(int *)(buf)=padcnt;
     cipe_prnpad(buf+sizeof(int), blockSize-sizeof(int));
-    ecb_enc(buf, buf, padkey);
+    ecb_enc((part*)buf, (part*)buf, &padkey);
 }
 #endif
 
@@ -167,11 +167,9 @@ void cipe_nodynkey(struct cipe *c)
 	     c->dev->name);
 }
 
+#if ProtocolVersion == 3
 
-#if ProtocolVersion == 3 || ProtocolVersion == d
-
-
-/* Encryption/decryption version 3 and d */
+/* Encryption/decryption version 3 */
 
 void cipe_encrypt(struct cipe *c, unsigned char *buf, int *len, int typ)
 {
@@ -188,8 +186,8 @@ void cipe_encrypt(struct cipe *c, unsigned char *buf, int *len, int typ)
     *((unsigned long *)(buf+(*len)-4))=
 	htonl(crc32(buf+blockSize, (*len)-blockSize-4));
 
-    dprintk3(DEB_CRYPT, KERN_INFO, "%s: encrypt pad %d len %d\n",
-	     c->dev->name, p, *len);
+    dprintk(DEB_CRYPT, (KERN_INFO "%s: encrypt typ %d pad %d len %d\n",
+                        c->dev->name, typ, p, *len));
     cbc_enc(buf, *len, c->haveskey ? &c->skey_e : &c->key_e);
     cipe_checkskey(c);
 }
@@ -213,8 +211,8 @@ unsigned short cipe_decrypt(struct cipe *c, unsigned char *buf, int *len)
     (*len)-=(p>>4)&7;
     cipe_checkrkey(c);
 #define CTLBITS 0x06
-    dprintk3(DEB_CRYPT, KERN_INFO, "%s: decrypt pad %d typ %02X\n",
-	     c->dev->name, (p>>4)&7, p&CTLBITS);
+    dprintk(DEB_CRYPT, (KERN_INFO "%s: decrypt len=%d pad=%d typ=%02X\n",
+                        c->dev->name, (*len), (p>>4)&7, p&CTLBITS));
     return (p&CTLBITS);
 }
 
