@@ -10,7 +10,7 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: ioctl.c,v 1.1.4.1 1998/11/14 21:14:16 olaf Exp $ */
+/* $Id: ioctl.c,v 1.3 1998/11/14 20:48:02 olaf Exp $ */
 
 #include "cipe.h"
 #include "ioctl.h"
@@ -48,3 +48,14 @@ int ioctl_attach(int fd, char *ifname, struct siocsifcipatt *sio)
     return ioctl(fd, SIOCSIFCIPATT, &ifr);
 }
 
+int ioctl_alloc(int fd, char *ifname, struct siocsifcipall *sio)
+{
+    IFRSETUP;
+    return ioctl(fd, SIOCSIFCIPALL, &ifr);
+}
+
+int ioctl_unalloc(int fd, char *ifname, struct siocsifcipall *sio)
+{
+    IFRSETUP;
+    return ioctl(fd, SIOCSIFCIPUNA, &ifr);
+}
