@@ -10,7 +10,7 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: ioctl.h,v 1.1 1996/10/05 23:59:24 olaf Exp $ */
+/* $Id: ioctl.h,v 1.2 1998/08/22 16:05:51 olaf Exp $ */
 
 #include "cipe.h"
 
@@ -18,3 +18,5 @@ extern int ioctl_getpar(int fd, char *ifname, struct siocgifcippar *sio);
 extern int ioctl_setpar(int fd, char *ifname, struct siocsifcippar *sio);
 extern int ioctl_setkey(int fd, char *ifname, struct siocsifcipkey *sio);
 extern int ioctl_attach(int fd, char *ifname, struct siocsifcipatt *sio);
+extern int ioctl_alloc(int fd, char *ifname, struct siocsifcipall *sio);
+extern int ioctl_unalloc(int fd, char *ifname, struct siocsifcipall *sio);
